#include <math.h>
#include <arm_neon.h>
#include <stdio.h>

/// Computes the sine and cosine of two angles
/// in: angles = Two angles, expressed in radians, in the [-PI,PI] range.
/// out: results = vector containing [sin(angles[0]),cos(angles[0]),sin(angles[1]),cos(angles[1])]
static inline void vsincos(const float angles[2], float results[4]) 
{
    static const float constants[]  = 
    { 
    /* q1 */  0,                M_PI_2,           0,                M_PI_2,
    /* q2 */  M_PI,             M_PI,             M_PI,             M_PI,
    /* q3 */  4.f/M_PI,         4.f/M_PI,         4.f/M_PI,         4.f/M_PI,
    /* q4 */ -4.f/(M_PI*M_PI), -4.f/(M_PI*M_PI), -4.f/(M_PI*M_PI), -4.f/(M_PI*M_PI),
    /* q5 */  2.f,              2.f,              2.f,              2.f,
    /* q6 */  .225f,            .225f,            .225f,            .225f
    };  
    asm volatile(
        // Load q0 with [angle1,angle1,angle2,angle2]
        "vldmia %1, { d3 }\n\t"
        "vdup.f32 d0, d3[0]\n\t"
        "vdup.f32 d1, d3[1]\n\t"
        // Load q1-q6 with constants
        "vldmia %2, { q1-q6 }\n\t"
        // Cos(x) = Sin(x+PI/2), so
        // q0 = [angle1, angle1+PI/2, angle2, angle2+PI/2]
        "vadd.f32 q0,q0,q1\n\t"
        // if angle1+PI/2>PI, substract 2*PI
        // q0-=(q0>PI)?2*PI:0
        "vcge.f32 q1,q0,q2\n\t"
        "vand.f32 q1,q1,q2\n\t"
        "vmls.f32 q0,q1,q5\n\t"
        // q0=(4/PI)*q0 - q0*abs(q0)*4/(PI*PI)
        "vabs.f32 q1,q0\n\t"
        "vmul.f32 q1,q0,q1\n\t"
        "vmul.f32 q0,q0,q3\n\t"
        "vmul.f32 q1,q1,q4\n\t"
        "vadd.f32 q0,q0,q1\n\t"
        // q0+=.225*(q0*abs(q0) - q0)
        "vabs.f32 q1,q0\n\t"
        "vmul.f32 q1,q0,q1\n\t"
        "vsub.f32 q1,q0\n\t"
        "vmla.f32 q0,q1,q6\n\t"
        "vstmia %0, { q0 }\n\t"
        :: "r"(results), "r"(angles), "r"(constants)
        : "memory","cc","q0","q1","q2","q3","q4","q5","q6"
    );  
}


static inline unsigned int get_cyclecount (void)
{
  unsigned int value;
  asm volatile ("MRC p15, 0, %0, c9, c13, 0\t\n": "=r"(value));  
  return value;
}

void init_perfcounters (int32_t do_reset, int32_t enable_divider)
{
  // in general enable all counters (including cycle counter)
  int32_t value = 1;

  // peform reset:  
  if (do_reset)
  {
    value |= 2;         // reset all counters to zero.
    value |= 4;         // reset cycle counter to zero.
  } 
  
  if (enable_divider)
    value |= 8;         // enable "by 64" divider for CCNT.
    
  value |= 16;
  
  // program the performance-counter control-register:
  asm volatile ("MCR p15, 0, %0, c9, c12, 0\t\n" :: "r"(value));  
  
  // enable all counters:  
  asm volatile ("MCR p15, 0, %0, c9, c12, 1\t\n" :: "r"(0x8000000f));  

  // clear overflows:
  asm volatile ("MCR p15, 0, %0, c9, c12, 3\t\n" :: "r"(0x8000000f));
}
   
  

float q[2*1024];
float result_asm  [4*1024];
float result_libc [4*1024];

void main (int argc, char **args)
{
  int i;
  
  unsigned t1, t2;
  int n;
  FILE *f;
  
  for (i=0; i<1024; i++)
  {
    q[i*2+0] = (float)i / 1024.0f * 3.14f;
    q[i*2+1] = -(float)i / 1024.0f * 3.14f;
  }

  init_perfcounters (1, 0);
 
  for (n=0; n<10; n++)
  {
    t1 = get_cyclecount();
    for (i=0; i<1024; i++)
      vsincos( &q[i*2], &result_asm[i*4]);
      
    t1 = get_cyclecount()-t1;

    t2 = get_cyclecount();
    for (i=0; i<1024; i++)
    {
      result_libc[i*4+0] = sin(q[i*2+0]);
      result_libc[i*4+1] = cos(q[i*2+0]);
      result_libc[i*4+2] = sin(q[i*2+1]);
      result_libc[i*4+3] = cos(q[i*2+1]);
    }
    t2 = get_cyclecount()-t2;
    
    printf ("asm-version %d, c-version %d\n", t1, t2);
  }

  f = fopen ("result.txt", "w");
  
  for (i=0; i<4096; i++)
    fprintf (f, "%2.8e %2.8e\n", result_asm[i], result_libc[i]);
    
  fclose (f);
    
}

